import asyncio
import concurrent.futures

from .RuntimeCtx import RuntimeCtx
from .runtime_ctx_settings import inject_alias, set_default_runtime_ctx

class Listener(object):
    def __init__(self):
        self.handlers = {}
        return

    def on(self, event, handler):
        if event not in self.handlers:
            self.handlers[event] = []
        self.handlers[event].append(handler)
        return self

    async def emit(self, event, *args, **kwargs):
        if event in self.handlers:
            for handler in self.handlers[event]:
                if asyncio.iscoroutinefunction(handler):
                    await handler(*args, **kwargs)
                else:
                    handler(*args, **kwargs)

class Session(object):
    def __init__(self, agent):
        self.agent_runtime_ctx = agent.runtime_ctx
        self.runtime_ctx = RuntimeCtx(agent.runtime_ctx)
        self.workflows = agent.workflows
        self.work_nodes = agent.work_nodes
        inject_alias(self, "agent")
        inject_alias(self, "session")
        inject_alias(self, "request")
        set_default_runtime_ctx(self.runtime_ctx, "session")
        self.request_runtime_ctx = RuntimeCtx(self.runtime_ctx)        
        set_default_runtime_ctx(self.request_runtime_ctx, "request")

        self.is_working = False
        self.listener = Listener()
        self.final_result = None
        return

    def set(self, key, value, **kwargs):
        target_runtime_ctx = kwargs.get("target_runtime_ctx", "session")
        if target_runtime_ctx == "request":
            self.request_runtime_ctx.set(key, value, **kwargs)
        else:
            self.runtime_ctx.set(key, value, **kwargs)
        return self

    def get(self, key, **kwargs):
        target_runtime_ctx = kwargs.get("target_runtime_ctx", "session")
        if target_runtime_ctx == "request":
            self.request_runtime_ctx.get(key, **kwargs)
        else:
            self.runtime_ctx.get(key, **kwargs)
        return self

    def append(self, key, value, **kwargs):
        target_runtime_ctx = kwargs.get("target_runtime_ctx", "session")
        if target_runtime_ctx == "request":
            self.request_runtime_ctx.append(key, value, **kwargs)
        else:
            self.runtime_ctx.append(key, value, **kwargs)  
        return self

    def extend(self, key, value, **kwargs):
        target_runtime_ctx = kwargs.get("target_runtime_ctx", "session")
        if target_runtime_ctx == "request":
            self.request_runtime_ctx.extend(key, value, **kwargs)
        else:
            self.runtime_ctx.extend(key, value, **kwargs)
        return self

    def on(self, event, handler):
        self.listener.on(event, handler)
        return self

    async def __start(self, workflow_name):
        if self.is_working:
            await asyncio.sleep(0.1)
            await self.__start(workflow_name)
        else:
            self.is_working = True
            workflow = self.workflows.get(workflow_name)
            if workflow:
                if not isinstance(workflow, list) or len(workflow) == 0:
                    raise Exception(f"[session start]: No work node in workflow { workflow_name }")

                self.final_result = None

                async def set_final_result(data):
                    self.final_result = data
                    return

                self.listener.on('done', set_final_result)

                async def reset(data):
                    self.request_runtime_ctx = None
                    self.request_runtime_ctx = RuntimeCtx(self.runtime_ctx) 
                    set_default_runtime_ctx(self.request_runtime_ctx, "request")
                    self.listener = None
                    self.listener = Listener()
                    self.is_working = False
                    return

                self.listener.on('workflow_finish', reset)

                for work_node_name in workflow:
                    work_node = self.work_nodes.get(work_node_name)
                    if asyncio.iscoroutinefunction(work_node["main"]):
                        await work_node["main"](\
                            self.request_runtime_ctx,\
                            listener = self.listener,\
                            process = work_node["process"],\
                            session_runtime_ctx = self.runtime_ctx,\
                            agent_runtime_ctx = self.agent_runtime_ctx,\
                            worker_agent = self.runtime_ctx.get("worker_agent")\
                        )
                    else:
                        work_node["main"](\
                            self.request_runtime_ctx,\
                            listener = self.listener,\
                            process = work_node["process"],\
                            session_runtime_ctx = self.runtime_ctx,\
                            agent_runtime_ctx = self.agent_runtime_ctx,\
                            worker_agent = self.runtime_ctx.get("worker_agent")\
                        )
                runtime_ctx_final_reply = self.request_runtime_ctx.get("final_reply")
                result = runtime_ctx_final_reply if runtime_ctx_final_reply else self.final_result
                await self.listener.emit('workflow_finish', result)
                return result
            else:
                raise Exception(f'[session start]: No workflow named { workflow_name }')

    def start(self, workflow_name="default"):
        old_loop = asyncio.get_event_loop()
        new_loop = asyncio.new_event_loop()
        asyncio.set_event_loop(new_loop)
        with concurrent.futures.ThreadPoolExecutor() as executor:
            future = executor.submit(asyncio.run, self.__start(workflow_name=workflow_name))
            result = future.result()
        new_loop.stop()
        asyncio.set_event_loop(old_loop)
        return result

    async def async_start(self, workflow_name="default"):
        return await self.__start(workflow_name=workflow_name)


import yaml
import collections

from .generate_prompt_process import update_process

def ordered_dict_representer(dumper, data):
    return dumper.represent_dict(data.items())

yaml.add_representer(collections.OrderedDict, ordered_dict_representer)

def generate_prompt(runtime_ctx, **kwargs):
    process = kwargs["process"]
    '''
    SYSTEM MESSAGE
    '''
    system_prompt = collections.OrderedDict()
    #generate role
    role = runtime_ctx.get("agent_role")
    use_agent_role = runtime_ctx.get("use_agent_role")
    if role and use_agent_role:
        system_prompt["ROLE SETTINGS"] = {}
        for role_key, role_desc in role.items():
            system_prompt["ROLE SETTINGS"].update({ role_key : str(role_desc) })
    #generate status
    status = runtime_ctx.get("agent_status")
    use_agent_status = runtime_ctx.get("use_agent_status")
    if status and use_agent_status:
        system_prompt["STATUS"] = {}
        for status_key, status_desc in status.items():
            system_prompt["STATUS"].update({ status_key : str(status_desc) })
    if len(system_prompt) > 0:
        runtime_ctx.set("request_system_message", { "role": "system", "content": yaml.dump(system_prompt, allow_unicode=True) })
    '''
    REQUEST MESSAGE
    '''
    request_prompt_dict = collections.OrderedDict()
    prompt_input = runtime_ctx.get("prompt_input")
    prompt_instruct = runtime_ctx.get("prompt_instruct")
    prompt_segments = runtime_ctx.get("prompt_segments")
    prompt_output = runtime_ctx.get("prompt_output")
    prompt_output_format = runtime_ctx.get("prompt_output_format")
    prompt_style = runtime_ctx.get("prompt_style")
    if prompt_input != None and prompt_instruct == None and prompt_output == None:
        runtime_ctx.set("request_prompt_message", { "role": "user", "content": str(prompt_input) })
        if runtime_ctx.get("is_debug"):
            print("[PROMPT]")
            print(str(prompt_input))
    else:
        if prompt_input:
            request_prompt_dict["INPUT"] = prompt_input
        if prompt_instruct:
            request_prompt_dict["INSTRUCTIONS"] = prompt_instruct
        if prompt_segments:
            runtime_ctx.set("use_segments", True)
            pass#<-todo: segments
        elif prompt_output:
            prompt_output_format = prompt_output_format.upper() if prompt_output_format else ""
            if prompt_output_format not in process["generate_output_format"]:
                if isinstance(prompt_output, (dict, list)):
                    prompt_output_format = "JSON"
                    runtime_ctx.set("prompt_output_format", "JSON")
                else:
                    prompt_output_format = "TEXT"
                    runtime_ctx.set("prompt_output_format", "TEXT")
            request_prompt_dict["OUTPUT REQUIREMENT"] = process["generate_output_format"][prompt_output_format](prompt_output)
        request_prompt = process["generate_prompt_structure"][prompt_style](request_prompt_dict)
        if runtime_ctx.get("is_debug"):
            print("[PROMPT]")
            print(request_prompt)
        runtime_ctx.set("request_prompt_message", { "role": "user", "content": request_prompt})
    return

def export(agently):
    agently\
        .manage_work_node("generate_prompt")\
        .set_main_func(generate_prompt)\
        .set_runtime_ctx({
            "use_agent_role": {
                "layer": "agent",
                "alias": { "set": "use_role" }
            },
            "agent_role": {
                "layer": "agent",
                "alias": { "set_kv": "set_role" ,"append_kv": "append_role" },
            },
            "use_agent_status": {
                "layer": "agent",
                "alias": { "set": "use_status" }
            },
            "agent_status": {
                "layer": "agent",
                "alias": { "set_kv": "set_status", "append_kv": "append_status" },
            },
            "prompt_input": {
                "layer": "request",
                "alias": { "set": "input" },
            },
            "prompt_instruct": {
                "layer": "request",
                "alias": { "set_kv": "instruct" },
            },
            "prompt_output": {
                "layer": "request",
                "alias": { "set": "output" },
            },
            "prompt_segments": {
                "layer": "request",
                "alias": { "append": "segments" }
            },
        })\
        .register()
    update_process(agently.manage_work_node("generate_prompt"))
    return


import asyncio
import aiohttp
import json
import uuid

from .loads_and_fix import loads_and_fix

#百度千帆大模型平台限时免费模型列表可查阅：
#https://cloud.baidu.com/doc/WENXINWORKSHOP/s/2lktec9ey

#获取千帆access_token，可复制下面的方法运行，有效期30天，过期后需要重新运行
'''
async def get_wx_access_token (api_key, secret_key):
    url = "https://aip.baidubce.com/oauth/2.0/token"
    data = {
        "grant_type": "client_credentials",
        "client_id": api_key,
        "client_secret": secret_key,
    }
    async with aiohttp.ClientSession() as session:
        async with session.post(\
            url,\
            data=data,\
        ) as response:
            response = await response.json()
            return response
result = asyncio.run(get_wx_access_token("Your-API-Key", "Your-Secret-Key"))
print(result["access_token"])
'''

'''
Baidu WenXin Workshop
'''

def prepare_request_data(runtime_ctx):
    wx_model_type = runtime_ctx.get("wx_model_type")
    wx_model_name = runtime_ctx.get("wx_model_name")
    #default request_data
    request_data = {
        "llm_url": f"https://aip.baidubce.com/rpc/2.0/ai_custom/v1/wenxinworkshop/{ wx_model_type }/{ wx_model_name }",
        "data": { "user_id": uuid.uuid4() },
    }
    #check auth info
    llm_auth = runtime_ctx.get("llm_auth")
    if "wenxin" in llm_auth:
        request_data.update({ "llm_auth": llm_auth["wenxin"] })
    else:
        raise Exception(f"[request]: llm_auth must be set to agent. use agent.set_llm_auth('wenxin', <Your-access-token>) to set.\nHow to generate access token: https://cloud.baidu.com/doc/WENXINWORKSHOP/s/Sllyztytp")
    #update user customize settings
    cus_llm_url = runtime_ctx.get("llm_url")
    cus_llm_url = cus_llm_url["wenxin"] if cus_llm_url and "wenxin" in cus_llm_url else None
    if cus_llm_url:
        request_data.update({ "llm_url": cus_llm_url })
    request_data.update({ "proxy": runtime_ctx.get("proxy") })
    cus_request_options = runtime_ctx.get("request_options")
    cus_request_options = cus_request_options["wenxin"] if cus_request_options and "wenxin" in cus_request_options else None
    if isinstance(cus_request_options, dict):
        for options_name, options_value in cus_request_options.items():
            request_data["data"].update({ options_name: options_value })
    #get request messages
    request_messages = runtime_ctx.get("request_messages")
    #transform messages format
    for message in request_messages:
        if message["role"] == "system":
            message["role"] = "user"
    request_data["data"].update({ "messages": request_messages })
    #prepare prompt for completions
    if wx_model_type == "completions":
        prompt = ""
        for message in request_messages:
            prompt += message["content"] + '\n'
        request_data["prompt"] = prompt
    return request_data

async def request(request_data, listener):
    url = request_data["llm_url"] + '?access_token=' + request_data["llm_auth"]
    data = request_data["data"].copy()
    if "prompt" in request_data:
        del data["messages"]
        data["prompt"] = request_data["prompt"]
    data["stream"] = False
    headers = {
        "Content-Type": "application/json",
        "Accept": "application/json",
    }
    proxy = request_data["proxy"] if "proxy" in request_data else {}
    async with aiohttp.ClientSession() as session:
        async with session.post(url, headers=headers, json=data, proxy=proxy) as response:
            response = await response.json()
            await listener.emit("response:done", response)
            return response

async def streaming(request_data, listener):
    url = request_data["llm_url"] + '?access_token=' + request_data["llm_auth"]
    data = request_data["data"].copy()
    if "prompt" in request_data:
        del data["messages"]
        data["prompt"] = request_data["prompt"]
    data["stream"] = True
    headers = {
        "Content-Type": "application/json",
        "Accept": "text/event-stream",
    }
    proxy = request_data["proxy"] if "proxy" in request_data else {}
    async with aiohttp.ClientSession() as session:
        async with session.post(url, headers=headers, json=data, proxy=proxy) as response:
            async for chunk in response.content.iter_chunks():
                if chunk[0]:
                    try:
                        delta = chunk[0].decode('utf-8')[6:][:-2]
                        delta = json.loads(delta)
                        await listener.emit("response:delta", delta)
                        if delta["is_end"]:
                            await listener.emit("response:done", None)
                    except Exception as e:
                        print(chunk[0])
                        error = json.loads(chunk[0])
                        if error["error_code"]:
                            raise Exception(error)

async def handle_response(listener, runtime_ctx, worker_agent):
    buffer = {
        "message": { "role": "assistant", "content": "" },
    }    
    async def handle_response_delta(delta_data):
        await listener.emit("extract:delta_full", delta_data)
        if "result" in delta_data:
            await listener.emit("extract:delta", delta_data["result"])
            buffer["message"]["content"] += delta_data["result"]
        return

    listener.on("response:delta", handle_response_delta)

    async def handle_response_done(done_data):
        if done_data != None:
            await listener.emit("extract:done_full", done_data)
            content = done_data["result"]
            prompt_output_format = runtime_ctx.get("prompt_output_format")
            content = await loads_and_fix(content, prompt_output_format, worker_agent=worker_agent, is_debug=runtime_ctx.get("is_debug"))
            await listener.emit("extract:done", content)
        else:
            await listener.emit("extract:done_full", buffer)
            await listener.emit("extract:done", buffer["message"]["content"])

    listener.on("response:done", handle_response_done)


def assemble_request_messages(runtime_ctx, **kwargs):
    session_runtime_ctx = kwargs["session_runtime_ctx"]

    request_system_message = runtime_ctx.get("request_system_message")
    request_context_messages = session_runtime_ctx.get("request_context_messages")
    request_prompt_message = runtime_ctx.get("request_prompt_message")
    request_messages = []
    if request_system_message:
        request_messages.append(request_system_message)
    if request_context_messages:
        request_messages.extend(request_context_messages)
    request_messages.append(request_prompt_message)
    runtime_ctx.set("request_messages", request_messages)

def export(agently):
    agently\
        .manage_work_node("assemble_request_messages")\
        .set_main_func(assemble_request_messages)\
        .set_runtime_ctx({
            "request_messages": { "default": [], "layer": "request" }
        })\
        .register()
    return


import asyncio
import json
import yaml
import re

async def loads_and_fix(content, output_format, fix_count=0, worker_agent=None, is_debug=False):
    fix_count += 1
    match = re.match(r"^```(.*)\n", content)
    if match:
        prefix = match.group(0)
        content = content.replace(prefix, "")
        content = content.replace("```", "")
    if output_format.upper() not in ("JSON", "YAML"):
        return content
    if worker_agent:
        async def fix_yaml_format(content, e):
            if is_debug:
                print('[Worker Agent Activated]')
                print('Fix YAML Format Error:', e)
                print('Current Ouput: ', content)
                print('\n')
            fixed_content = await worker_agent\
                .input({
                    "origin YAML String": content,
                    "error": e,
                })\
                .output('Fixed YAML String only without explanation and decoration.')\
                .async_start()
            if is_debug:
                print('Fixed Content:', fixed_content)
                print('\n')
            return fixed_content

        if fix_count >= 3:
            raise Exception(f'Format Error can not be fixed: [{ output_format }]\n{ content }')
        try:
            if output_format.upper() == "JSON":
                content = json.loads(content)
                return content
            elif output_format.upper() == "YAML":
                content = yaml.safe_load(content)
                prompt_output = runtime_ctx.get("prompt_output")
                if len(content.keys()) < len(prompt_output.keys()):
                    if is_debug:
                        print('[Worker Agent Activated]')
                        print('Fix YAML Format Error: Not enough output keys.')
                        print('Output requirement: ', str(runtime_ctx.get("prompt_output")))
                        print('Current Output: ', content)
                        print('\n')
                    fixed_content = await worker_agent\
                        .input({
                            "output keys requirement": str(runtime_ctx.get("prompt_output")),
                            "current output": content
                        })\
                        .output('Fixed YAML FORMAT STRING ONLY without explanation and decoration.')\
                        .async_start()
                    if is_debug:
                        print('Fixed Content:', fixed_content)
                        print('\n')
                    return await loads_and_fix(fixed_content, output_format, fix_count)
                else:
                    return content
        except json.JSONDecodeError as e:
            if is_debug:
                print('[Worker Agent Activated]')
                print('Fix JSON Format Error:', e.msg)
                print('Current Ouput: ', content)
                print('\n')
            fixed_content = await worker_agent\
                .input({
                    "origin JSON String": content,
                    "error": e.msg,
                    "position": e.pos,
                })\
                .output('Fixed JSON String only without explanation and decoration.')\
                .async_start()
            if is_debug:
                print('Fixed Content:', fixed_content)
                print('\n')
            return await loads_and_fix(fixed_content, output_format, fix_count=fix_count, worker_agent=worker_agent, is_debug=is_debug)
        except yaml.scanner.ScannerError as e:
            fixed_content = await fix_yaml_format(content, e)
            return await loads_and_fix(fixed_content, output_format, fix_count=fix_count, worker_agent=worker_agent, is_debug=is_debug)
        except yaml.parser.ParserError as e:
            fixed_content = await fix_yaml_format(fixed_content, e)
            return await loads_and_fix(fixed_content, output_format, fix_count=fix_count, worker_agent=worker_agent, is_debug=is_debug)
    else:
        if output_format.upper() == "JSON":
            return json.loads(content)
        if output_format.upper() == "YAML":
            return yaml.safe_load(content)


import copy

class RuntimeCtx(object):
    def __init__ (self, father = None, init_data = {}, **kwargs):
        self.default_domain = kwargs.get("domain", "default")
        self.father = father
        self.runtime_ctx_data = init_data.copy()
        return

    def __update_dict(self, target, dict_item):
        for key in dict_item.keys():
            if key not in target:
                target[key] = {}
            if isinstance(dict_item[key], dict):
                self.__update_dict(target[key], dict_item[key])
            else:
                if dict_item[key] != None:
                    target[key] = dict_item[key]

    def __action(self, action_name, target, keys, value):
        if len(keys) == 1:
            if action_name == "set":
                target.update({ keys[0]: value })
            if action_name == "append":
                key = keys[0]
                if key not in target:
                    target[key] = [value]
                else:
                    target[key].append(value)
            if action_name == "extend":
                key = keys[0]
                if key not in target or not isinstance(target[key], list):
                    target[key] = value
                else:
                    target[key].extend(value)
            if action_name == "add":
                key = keys[0]
                if key not in target:
                    target[key] = set([value])
                else:
                    target[key].add(value)
            if action_name == "update":
                key = keys[0]
                if key not in target:
                    if value != None:
                        target[key] = value
                else:
                    if isinstance(value, dict):
                        self.__update_dict(target[key], value)
                    else:
                        if value != None:
                            target.update({ key: value })
        else:
            key = keys[0]
            if key not in target:
                target[key] = {}
            self.__action(action_name, target[key], keys[1:], value)
        return

    def _action(self, action_name, keys, value, **kwargs):
        domain = kwargs.get("domain", self.default_domain)
        if domain not in self.runtime_ctx_data:
            self.runtime_ctx_data[domain] = {}
        splited_keys = keys.split('.')
        value = copy.deepcopy(value)
        self.__action(action_name, self.runtime_ctx_data[domain], splited_keys, value)
        return

    def set(self, keys, value, **kwargs):
        self._action("set", keys, value, **kwargs)
        return self

    def append(self, keys, value, **kwargs):
        self._action("append", keys, value, **kwargs)
        return self

    def extend(self, keys, value, **kwargs):
        self._action("extend", keys, value, **kwargs)
        return self

    def add(self, keys, value, **kwargs):
        self._action("add", keys, value, **kwargs)
        return self

    def update(self, keys, value, **kwargs):
        self._action("update", keys, value, **kwargs)
        return self
    
    def set_all(self, data, **kwargs):
        domain = kwargs.get("domain", self.default_domain)
        data = copy.deepcopy(data)
        self.runtime_ctx_data.update({ domain: data })
        return self

    def clear(self, **kwargs):
        domain = kwargs.get("domain", self.default_domain)
        self.runtime_ctx_data.update({ domain: {} })
        return self

    def get(self, key, **kwargs):
        domain = kwargs.get("domain", self.default_domain)
        if domain in self.runtime_ctx_data and key in self.runtime_ctx_data[domain]:
            return copy.deepcopy(self.runtime_ctx_data[domain][key])
        else:
            return None if self.father == None else copy.deepcopy(self.father.get(key, domain = domain))

    def get_all(self, **kwargs):
        domain = kwargs.get("domain", self.default_domain)
        #return copy.deepcopy(self.runtime_ctx_data[domain]) if domain in self.runtime_ctx_data else None
        return copy.deepcopy(self.runtime_ctx_data[domain]) if domain in self.runtime_ctx_data else None

    def __get_father_list(self, target, current_list):
        if target.father != None:
            current_list.append(target)
            return self.__get_father_list(target.father, current_list)
        else:
            current_list.append(target)
            return current_list

    def get_all_above(self, **kwargs):
        domain = kwargs.get("domain", self.default_domain)
        father_list = self.__get_father_list(self, [])
        reverse_list = father_list[::-1]
        result = {}
        for target in father_list:
            target_all_dict = target.get_all(domain=domain)
            target_all_dict = target_all_dict if target_all_dict else {}
            self.__update_dict(result, target_all_dict)
        return copy.deepcopy(result)

    def get_all_domain (self):
        return copy.deepcopy(self.runtime_ctx_data)


import copy

from .runtime_ctx_settings import inject_alias, set_default_runtime_ctx
from .RuntimeCtx import RuntimeCtx
from .Session import Session

class Agent(object):
    def __init__(self, agently, blueprint = None):
        if blueprint:
            self.runtime_ctx = copy.deepcopy(blueprint.runtime_ctx)
            self.workflows = copy.deepcopy(blueprint.workflows)
            self.work_nodes = copy.deepcopy(blueprint.work_nodes)
        else:
            self.runtime_ctx = RuntimeCtx(agently.runtime_ctx)
            self.workflows = agently.workflows
            self.work_nodes = agently.work_nodes
        inject_alias(self, "agent")
        set_default_runtime_ctx(self.runtime_ctx, "agent")
        return
    '''
    Basic RuntimeCtx Management
    '''
    def set(self, key, value, **kwargs):
        self.runtime_ctx.set(key, value, **kwargs)
        return self

    def get(self, key, **kwargs):
        self.runtime_ctx.get(key, **kwargs)
        return self

    def append(self, key, value, **kwargs):
        self.runtime_ctx.append(key, value, **kwargs)
        return self

    def extend(self, key, value, **kwargs):
        self.runtime_ctx.extend(key, value, **kwargs)
        return self
    '''
    Create Session
    '''
    def create_session(self):
        return Session(self)


from .GPT import\
    prepare_request_data as prepare_request_data_gpt,\
    request as request_gpt,\
    streaming as streaming_gpt,\
    handle_response as handle_response_gpt

from .MiniMax import\
    prepare_request_data as prepare_request_data_minimax,\
    request as request_minimax,\
    streaming as streaming_minimax,\
    handle_response as handle_response_minimax

from .Spark import\
    prepare_request_data as prepare_request_data_spark,\
    request as request_spark,\
    streaming as streaming_spark,\
    handle_response as handle_response_spark

from .wenxin import\
    prepare_request_data as prepare_request_data_wenxin,\
    request as request_wenxin,\
    streaming as streaming_wenxin,\
    handle_response as handle_response_wenxin


from .RuntimeCtx import RuntimeCtx
from .WorkNodes import WorkNodes
from .Workflows import Workflows
from .Blueprint import Blueprint
from .Agent import Agent
from .EnvBox import EnvBox
from .work_nodes import init_worker_agent, manage_context, generate_prompt, assemble_request_messages, register_response_suffix, request

class Agently(object):
    def __init__(self):
        self.runtime_ctx = RuntimeCtx(None)
        self.work_nodes = WorkNodes(self.runtime_ctx)
        self.workflows = Workflows(self.runtime_ctx)
        return

    def manage_work_node(self, work_node_name):
        return self.work_nodes.manage_work_node(work_node_name)

    def set_workflow(self, workflow, *, name="default"):
        return self.workflows.set(name, workflow)

    def create_blueprint(self):
        return Blueprint(self)

    def create_agent(self, blueprint = None):
        return Agent(self, blueprint)

    def create_env_box(self):
        return EnvBox(self)

    def set(self, key, value, **kwargs):
        self.runtime_ctx.set(key, value, **kwargs)
        return self

    def get(self, key, **kwargs):
        self.runtime_ctx.get(key, **kwargs)
        return self

    def append(self, key, value, **kwargs):
        self.runtime_ctx.append(key, value, **kwargs)
        return self

    def extend(self, key, value, **kwargs):
        self.runtime_ctx.extend(key, value, **kwargs)
        return self

    def use(self, inject_funcs):
        if not isinstance(inject_funcs, list):
            if not callable(inject_funcs):
                raise Exception("[Agently use]: .use() must pass on a function list or a function.")
            else:
                inject_funcs = [inject_funcs]
        for inject_func in inject_funcs:
            if not callable(inject_func):
                raise Exception("[Agently use]: .use() must pass on a callable function.")
            else:
                inject_func(self)
        return self

def create_empty():
    return Agently()

def create_worker():
    new_agently = Agently()
    worker_agent_blueprint = new_agently.create_blueprint()
    worker_agent_blueprint.use([init_worker_agent, generate_prompt, assemble_request_messages, register_response_suffix, request])
    worker_agent_blueprint.set_workflow(["init_worker_agent", "generate_prompt", "assemble_request_messages", "register_response_suffix", "request"])
    worker_agent = new_agently.create_agent(worker_agent_blueprint)
    worker_agent.set("is_worker_agent", True)
    new_agently.set("worker_agent", worker_agent)
    return worker_agent.create_session()

def create():
    new_agently = Agently()
    new_agently.use([init_worker_agent, manage_context, generate_prompt, assemble_request_messages, register_response_suffix, request])
    new_agently.set_workflow(["init_worker_agent", "manage_context", "generate_prompt", "assemble_request_messages", "register_response_suffix", "request"])
    worker_agent = create_worker()
    new_agently.set("worker_agent", worker_agent)
    return new_agently


import asyncio

from .request_process import update_process

'''
Main
'''
async def request_main_default(runtime_ctx, **kwargs):
    process = kwargs["process"]
    listener = kwargs["listener"]
    worker_agent = runtime_ctx.get("worker_agent")
    llm_name = runtime_ctx.get("llm_name")

    #prepare request data
    request_data = process["prepare_request_data"][llm_name](runtime_ctx)

    if runtime_ctx.get("is_debug"):
        print("[REQUEST]")
        print(request_data["data"]["messages"])

    is_streaming = runtime_ctx.get("is_streaming")

    #register all handlers
    await process["handle_response"][llm_name](listener, runtime_ctx, worker_agent)
    
    #then start request    
    if not is_streaming:
        response = await process["request_llm"][llm_name](request_data, listener)
    else:
        response = await process["streaming_llm"][llm_name](request_data, listener)
    return

async def request(runtime_ctx, **kwargs):
    process = kwargs["process"]
    request_method = runtime_ctx.get("request_method")
    if request_method not in process["request_main"]:
        request_method = "default"
    await process["request_main"][request_method](runtime_ctx, **kwargs)
    return

def export(agently):
    agently\
        .manage_work_node("request")\
        .set_main_func(request)\
        .set_process("request_main", request_main_default, "default")\
        .set_runtime_ctx({
            "request_method": {
                "layer": "agent",
                "alias": { "set": "set_request_method" },
                "default": "default",
            },
            "llm_name": {
                "layer": "agent",
                "alias": { "set": "set_llm_name" },
                "default": "GPT"
            },
            "llm_url": {
                "layer": "agent",
                "alias": { "set_kv": "set_llm_url" },
            },
            "llm_auth": {
                "layer": "agent",
                "alias": { "set_kv": "set_llm_auth" },
            },
            "request_options": {
                "layer": "agent",
                "alias": { "set_kv": "set_request_options" },
            },
            "proxy": {
                "layer": "agent",
                "alias": { "set": "set_proxy" },
            },
            "is_streaming": {
                "layer": "session",
                "alias": { "set": "set_streaming" },
                "default": False,
            },
        })\
        .register()
    update_process(agently.manage_work_node("request"))
    return


import asyncio
async def init_worker_agent(runtime_ctx, **kwargs):
    is_worker_agent = runtime_ctx.get("is_worker_agent")
    worker_agent = kwargs["worker_agent"]
    if is_worker_agent:
        return
    elif worker_agent != None:
        llm_name = runtime_ctx.get("llm_name")
        llm_url = runtime_ctx.get("llm_url")
        llm_auth = runtime_ctx.get("llm_auth")
        proxy = runtime_ctx.get("proxy")
        request_options = runtime_ctx.get("request_options")

        # GPT first
        if "GPT" in llm_auth and llm_auth["GPT"] != None:
            worker_agent.set_llm_name("GPT")
            if llm_url and "GPT" in llm_url:
                worker_agent.set_llm_url("GPT", llm_url["GPT"])
            worker_agent.set_llm_auth("GPT", llm_auth["GPT"])
            if proxy:
                worker_agent.set_proxy(runtime_ctx.get("proxy"))
            if request_options and "GPT" in request_options:
                worker_agent.set_request_options("GPT", request_options["GPT"])
        # No GPT then use user current llm
        else:
            worker_agent.set_llm_name(llm_name)
            if llm_url and llm_name in llm_url:
                worker_agent.set_llm_url(llm_name, llm_url[llm_name])
            if llm_auth and llm_name in llm_auth:
                worker_agent.set_llm_auth(llm_name, llm_auth[llm_name])
            if proxy:
                worker_agent.set_proxy(runtime_ctx.get("proxy"))
            if request_options and llm_name in request_options:
                worker_agent.set_request_options(llm_name, request_options[llm_name])
        return

def export(agently):
    agently\
        .manage_work_node("init_worker_agent")\
        .set_main_func(init_worker_agent)\
        .register()
    return


import asyncio
import copy

import _thread as thread
import base64
import datetime
import hashlib
import hmac
import json
from urllib.parse import urlparse
import ssl
from datetime import datetime
from time import mktime
from urllib.parse import urlencode
from wsgiref.handlers import format_date_time

import websockets

from .loads_and_fix import loads_and_fix

'''
XunFei Spark
'''
class WsParam(object):
    # 初始化
    def __init__(self, APPID, APIKey, APISecret, Spark_url):
        self.APPID = APPID
        self.APIKey = APIKey
        self.APISecret = APISecret
        self.host = urlparse(Spark_url).netloc
        self.path = urlparse(Spark_url).path
        self.Spark_url = Spark_url

    # 生成url
    def create_url(self):
        # 生成RFC1123格式的时间戳
        now = datetime.now()
        date = format_date_time(mktime(now.timetuple()))

        # 拼接字符串
        signature_origin = "host: " + self.host + "\n"
        signature_origin += "date: " + date + "\n"
        signature_origin += "GET " + self.path + " HTTP/1.1"

        # 进行hmac-sha256进行加密
        signature_sha = hmac.new(self.APISecret.encode('utf-8'), signature_origin.encode('utf-8'),
                                 digestmod=hashlib.sha256).digest()

        signature_sha_base64 = base64.b64encode(signature_sha).decode(encoding='utf-8')

        authorization_origin = f'api_key="{self.APIKey}", algorithm="hmac-sha256", headers="host date request-line", signature="{signature_sha_base64}"'

        authorization = base64.b64encode(authorization_origin.encode('utf-8')).decode(encoding='utf-8')

        # 将请求的鉴权参数组合为字典
        v = {
            "authorization": authorization,
            "date": date,
            "host": self.host
        }
        # 拼接鉴权参数，生成url
        url = self.Spark_url + '?' + urlencode(v)
        # 此处打印出建立连接时候的url,参考本demo的时候可取消上方打印的注释，比对相同参数时生成的url与自己代码生成的url是否一致
        return url

def prepare_request_data(runtime_ctx):
    #default request_data
    request_data = {
        "llm_url": "wss://spark-api.xf-yun.com/v1.1/chat",
        "data": {
            "header": {
                "uid": "0000",
            },
            "parameter": {
                "chat": {
                    "domain": "general",
                    "temperature": 0.5,
                    "max_tokens": 4096,
                    "auditing": "default"
                }
            },
        },
    }
    #check auth info
    llm_auth = runtime_ctx.get("llm_auth")
    if "Spark" in llm_auth:
        request_data["data"]["header"].update({ "app_id": llm_auth["Spark"]["app_id"] })
        request_data.update({ "llm_auth": llm_auth["Spark"] })
    else:
        raise Exception("[request]: llm_auth must be set to agent. use agent.set_llm_auth('Spark', { 'app_id': '...', 'api_secret': '...' , 'api_key': '...' }) to set.")
    #update user customize settings
    cus_llm_url = runtime_ctx.get("llm_url")
    cus_llm_url = cus_llm_url["Spark"] if cus_llm_url and "Spark" in cus_llm_url else None
    if cus_llm_url:
        request_data.update({ "llm_url": cus_llm_url })
    request_data.update({ "proxy": runtime_ctx.get("proxy") })
    cus_request_options = runtime_ctx.get("request_options")
    cus_request_options = cus_request_options["Spark"] if cus_request_options and "Spark" in cus_request_options else None
    if isinstance(cus_request_options, dict):
        for options_name, options_value in cus_request_options.items():
            request_data["data"]["parameter"]["chat"].update({ options_name: options_value })
    #get request messages
    request_messages = runtime_ctx.get("request_messages")
    #transform messages format
    for message in request_messages:
        if message["role"] == "system":
            message["role"] = "user"
    request_data["data"].update({ "messages": request_messages })
    return request_data

async def request(request_data, listener):
    response_message = { "content": "" }
    request_data["data"]["payload"] = {
        "message": {
            "text": copy.deepcopy(request_data["data"]["messages"])
        }
    }
    del request_data["data"]["messages"]

    llm_auth = request_data["llm_auth"]
    llm_url = request_data["llm_url"]
    ws_param = WsParam(llm_auth["app_id"], llm_auth["api_key"], llm_auth["api_secret"], llm_url)
    ws_url = ws_param.create_url()
    async with websockets.connect(ws_url) as websocket:
        await websocket.send(json.dumps(request_data["data"]))
        async for message in websocket:
            data = json.loads(message)
            code = data["header"]["code"]
            if code != 0:
                raise Exception("[Request Spark]:", code, data)
            else:
                choices = data["payload"]["choices"]
                await listener.emit("response:delta", choices)
                status = choices["status"]
                content = choices["text"][0]["content"]        
                response_message["content"] = response_message["content"] + content
                if status == 2:
                    del choices["text"][0]["content"]
                    response_message.update(choices["text"][0])
                    await listener.emit('response:done', response_message)       
                    return response_message

async def streaming(request_data, listener):
    return await request(request_data, listener)

async def handle_response(listener, runtime_ctx, worker_agent):
    is_streaming = runtime_ctx.get("is_streaming")
    prompt_output_format = runtime_ctx.get("prompt_output_format")
    
    async def handle_response_delta(delta_data):
        await listener.emit("extract:delta_full", delta_data)
        await listener.emit("extract:delta", delta_data["text"][0]["content"])

    listener.on("response:delta", handle_response_delta)

    async def handle_response_done(done_data):
        await listener.emit("extract:done_full", done_data)
        prompt_output_format = runtime_ctx.get("prompt_output_format")
        content = await loads_and_fix(done_data["content"], prompt_output_format, worker_agent=worker_agent, is_debug=runtime_ctx.get("is_debug"))
        await listener.emit("extract:done", content)

    listener.on("response:done", handle_response_done)


from .llm_request import\
    prepare_request_data_gpt, request_gpt, streaming_gpt, handle_response_gpt,\
    prepare_request_data_minimax, request_minimax, streaming_minimax, handle_response_minimax,\
    prepare_request_data_spark, request_spark, streaming_spark, handle_response_spark,\
    prepare_request_data_wenxin, request_wenxin, streaming_wenxin, handle_response_wenxin

def update_process(work_node_management):
    result = work_node_management\
        .set_process("prepare_request_data", prepare_request_data_gpt, "GPT")\
        .set_process("request_llm", request_gpt, "GPT")\
        .set_process("streaming_llm", streaming_gpt, "GPT")\
        .set_process("handle_response", handle_response_gpt, "GPT")\
        .set_process("prepare_request_data", prepare_request_data_minimax, "MiniMax")\
        .set_process("request_llm", request_minimax, "MiniMax")\
        .set_process("streaming_llm", streaming_minimax, "MiniMax")\
        .set_process("handle_response", handle_response_minimax, "MiniMax")\
        .set_process("prepare_request_data", prepare_request_data_spark, "Spark")\
        .set_process("request_llm", request_spark, "Spark")\
        .set_process("streaming_llm", streaming_spark, "Spark")\
        .set_process("handle_response", handle_response_spark, "Spark")\
        .set_process("prepare_request_data", prepare_request_data_wenxin, "wenxin")\
        .set_process("request_llm", request_wenxin, "wenxin")\
        .set_process("streaming_llm", streaming_wenxin, "wenxin")\
        .set_process("handle_response", handle_response_wenxin, "wenxin")\
        .set_runtime_ctx({
            "minimax_bot_name": {
                "layer": "agent",
                "alias": { "set": "set_minimax_bot_name" },
                "default": "BOT",
            },
            "minimax_user_name": {
                "layer": "agent",
                "alias": { "set": "set_minimax_user_name" },
                "default": "USER",
            },
            "wx_model_name": {
                "layer": "agent",
                "alias": { "set": "set_wx_model_name" },
                "default": "qianfan_chinese_llama_2_7b",
            },
            "wx_model_type": {
                "layer": "agent",
                "alias": { "set": "set_wx_model_type" },
                "default": "chat", #"chat" | "completions"
            },
        })\
        .update()
    return


from .RuntimeCtx import RuntimeCtx

class Workflows(object):
    def __init__(self, runtime_ctx):
        self.runtime_ctx = runtime_ctx
        return

    def set(self, workflow_name, workflow):
        error_list = []
        if not isinstance(workflow_name, str):
            error_list.append("[set_workflow]: workflow_name must be a string.")
        if not isinstance(workflow, list):
            error_list.append("[set_workflow]: workflow must be a list.")
        if len(error_list) > 0:
            return {
                "ok": False,
                "message": {
                    "error": error_list,
                },
            }
        work_nodes = self.runtime_ctx.get_all_above(domain = "work_nodes")
        for work_node_name in workflow:
            if work_node_name not in work_nodes:
                error_list.append(f"[set_workflow]: '{ work_node_name }' is not in work_nodes dict.")
        if len(error_list) > 0:
            return {
                "ok": False,
                "message": {
                    "error": error_list,
                },
            }
        self.runtime_ctx.set(workflow_name, workflow, domain = "workflows")
        return { "ok": True }

    def get(self, workflow_name):
        return self.runtime_ctx.get(workflow_name, domain = "workflows")


def __inject_alias(targetClass, runtime_ctx_name, alias_dict, work_node_name, layer):
    for alias_type, alias_name in alias_dict.items():
        if alias_type == "get":
            setattr(targetClass, alias_name, lambda *, runtime_ctx_name=runtime_ctx_name: targetClass.get(runtime_ctx_name, target_runtime_ctx=layer))
        if alias_type == "set":
            setattr(targetClass, alias_name, lambda value, *, runtime_ctx_name=runtime_ctx_name: targetClass.set(runtime_ctx_name, value, target_runtime_ctx=layer))
        if alias_type == "get_k":
            setattr(targetClass, alias_name, lambda key, *, runtime_ctx_name=runtime_ctx_name: targetClass.runtime_ctx.get(runtime_ctx_name, target_runtime_ctx=layer)[key])
        if alias_type == "set_kv":
            setattr(targetClass, alias_name, lambda item_key, item_value, *, runtime_ctx_name=runtime_ctx_name: targetClass.set(f"{ runtime_ctx_name }.{ item_key }", item_value, target_runtime_ctx=layer))
        if alias_type == "append":
            setattr(targetClass, alias_name, lambda value, *, runtime_ctx_name=runtime_ctx_name: targetClass.append(runtime_ctx_name, value, target_runtime_ctx=layer))
        if alias_type == "append_kv":
            setattr(targetClass, alias_name, lambda item_key, item_value, *, runtime_ctx_name=runtime_ctx_name: targetClass.append(f"{ runtime_ctx_name }.{ item_key }", item_value, target_runtime_ctx=layer))
        if alias_type == "extend":
            setattr(targetClass, alias_name, lambda value, *, runtime_ctx_name=runtime_ctx_name: targetClass.extend(runtime_ctx_name, value, target_runtime_ctx=layer))
        if alias_type == "customize":
            alias_func = alias_name["func"]
            alias_name = alias_name["name"]
            setattr(targetClass, alias_name, alias_func)
        alias_record = { alias_name: { "type": alias_type, "runtime_ctx_name": runtime_ctx_name, "work_node_name": work_node_name } }
        alias_records = targetClass.runtime_ctx.get("alias_records")
        alias_records = alias_records if alias_records else {}
        if alias_name in alias_records:
            if alias_records[alias_name]["type"] != alias_type or alias_records[alias_name]["runtime_ctx_name"] != runtime_ctx_name:
                raise Exception({
                    "err": "alias conflict",
                    "alias name": alias_records[alias_name],
                    "alias record": alias_record,
                })
        else:
            targetClass.runtime_ctx.update("alias_records", alias_record)
    return


def inject_alias(targetClass, layer):
    work_nodes = targetClass.runtime_ctx.get_all_above(domain = "work_nodes")
    if work_nodes == None:
        raise Exception("[init agent runtime_ctx]: No work node is stated. Use Agently/Blueprint.manage_work_node(<work_node_name>) to create one first.")
    for work_node_name, work_node in work_nodes.items():
        if "runtime_ctx_settings" in work_node:
            runtime_ctx_settings = work_node["runtime_ctx_settings"]
            for runtime_ctx_name in runtime_ctx_settings.keys():
                runtime_ctx_setting = runtime_ctx_settings[runtime_ctx_name]
                #confirm layer
                if "layer" in runtime_ctx_setting and runtime_ctx_setting["layer"] == layer:
                    #inject alias
                    if "alias" in runtime_ctx_setting:
                        __inject_alias(targetClass, runtime_ctx_name, runtime_ctx_setting["alias"], work_node_name, layer)

def set_default_runtime_ctx(target_runtime_ctx, layer):
    work_nodes = target_runtime_ctx.get_all_above(domain = "work_nodes")
    if work_nodes == None:
        raise Exception("[init agent runtime_ctx]: No work node is stated. Use Agently/Blueprint.manage_work_node(<work_node_name>) to create one first.")
    for work_node_name, work_node in work_nodes.items():
        if "runtime_ctx_settings" in work_node:
            runtime_ctx_settings = work_node["runtime_ctx_settings"]
            for runtime_ctx_name in runtime_ctx_settings.keys():
                runtime_ctx_setting = runtime_ctx_settings[runtime_ctx_name]
                #confirm layer
                if "layer" in runtime_ctx_setting and runtime_ctx_setting["layer"] == layer:
                    #set default value
                    if "default" in runtime_ctx_setting:
                        target_runtime_ctx.set(runtime_ctx_name, runtime_ctx_setting["default"])


from .RuntimeCtx import RuntimeCtx

class WorkNodeManagement(object):
    def __init__(self, work_nodes, work_node_name):
        self.work_node_name = work_node_name
        self.runtime_ctx = RuntimeCtx()
        self.work_nodes = work_nodes
        self.__init_runtime_ctx()
        return

    def __init_runtime_ctx(self):
        self.runtime_ctx.set("main", None)
        self.runtime_ctx.set("process", {})
        self.runtime_ctx.set("runtime_ctx_settings", {})
        self.runtime_ctx.set("error_list", [])
        self.runtime_ctx.set("warning_list", [])
        return

    def set_main_func(self, main_func):
        if not callable(main_func):
            self.runtime_ctx.append("error_list", "[set_main_func]: main_func must be a callable function.")
        else:
            self.runtime_ctx.set("main", main_func)            
        return self

    def set_process(self, process_name, process_func, process_condition = "default"):
        if not isinstance(process_name, str):
            self.runtime_ctx.append("error_list", "[set_process]: process_name must be a string.")
        elif not callable(process_func):
            self.runtime_ctx.append("error_list", "[set_process]: process_func must be a callable function.")
        else:
            process_runtime_ctx = self.runtime_ctx.get("process")
            if process_name not in process_runtime_ctx:
                process_runtime_ctx.update({ process_name: {} })
            process_runtime_ctx[process_name].update({ process_condition: process_func })
            self.runtime_ctx.set("process", process_runtime_ctx)
        return self

    def set_runtime_ctx(self, runtime_ctx_settings_dict):
        self.runtime_ctx.update("runtime_ctx_settings", runtime_ctx_settings_dict)
        return self

    def __check(self, data):
        if data["main"] == None:
            self.runtime_ctx.append("error_list", "[register]: No main func is set. Use .set_main_func(main_func) to set.")
        if len(data["process"]) == 0:
            self.runtime_ctx.append("warning_list", "[register]: No process is set. Use .set_process(process_name, process_func, <process_condition>) to set if needed.")
        if len(data["runtime_ctx_settings"]) == 0:
            self.runtime_ctx.append("warning_list", "[register]: No runtime_ctx_settings is set. Use .set_runtime_ctx(runtime_ctx_settings_dict) to set if needed.")
        return 

    def register(self):
        runtime_ctx_data = {
            "main": self.runtime_ctx.get("main"),
            "process": self.runtime_ctx.get("process"),
            "runtime_ctx_settings": self.runtime_ctx.get("runtime_ctx_settings"),
        }

        self.__check(runtime_ctx_data)
        if self.work_nodes.get(self.work_node_name) != None:
            self.runtime_ctx.append("error_list", f"[register]: Work node '{ self.work_node_name }' existed.")

        error_list = self.runtime_ctx.get("error_list")
        warning_list = self.runtime_ctx.get("warning_list")

        if len(error_list) == 0:
            self.work_nodes.set(\
                self.work_node_name,\
                runtime_ctx_data\
            )
            self.__init_runtime_ctx()
            return {
                "ok": True,
                "message": {
                    "error": error_list,
                    "warning": warning_list
                }
            }
        else:
            self.__init_runtime_ctx()
            raise Exception({
                "ok": False,
                "message": {
                    "error": error_list,
                    "warning": warning_list,
                },
            })

    def __update_dict(self, target, dict_item):
        for key in dict_item.keys():
            if key not in target:
                target[key] = {}
            if isinstance(dict_item[key], dict):
                self.__update_dict(target[key], dict_item[key])
            else:
                if dict_item[key] != None:
                    target[key] = dict_item[key]

    def update(self):
        work_node = self.work_nodes.get(self.work_node_name)

        if work_node == None:
            raise Exception({
                "ok": False,
                "message": {
                    "error": [f"[update]: Work node '{ self.work_node_name }' is not existed."],
                    "warning": [],
                },
            })

        update_work_node_dict = {
            "main": self.runtime_ctx.get("main"),
            "process": self.runtime_ctx.get("process"),
            "runtime_ctx_settings": self.runtime_ctx.get("runtime_ctx_settings"),
        }

        self.__update_dict(work_node, update_work_node_dict)

        self.__check(work_node)

        error_list = self.runtime_ctx.get("error_list")
        warning_list = self.runtime_ctx.get("warning_list")

        if len(error_list) == 0:
            self.work_nodes.set(\
                self.work_node_name,\
                work_node\
            )
            self.__init_runtime_ctx()
            return {
                "ok": True,
                "message": {
                    "error": error_list,
                    "warning": warning_list
                }
            }
        else:
            self.__init_runtime_ctx()
            raise Exception({
                "ok": False,
                "message": {
                    "error": error_list,
                    "warning": warning_list,
                },
            })


class WorkNodes(object):
    def __init__(self, runtime_ctx):
        self.runtime_ctx = runtime_ctx
        return

    def manage_work_node(self, work_node_name):
        return WorkNodeManagement(self, work_node_name)

    def set(self, work_node_name, work_node):
        self.runtime_ctx.set(work_node_name, work_node, domain = "work_nodes")
        return

    def update(self, work_node_name, update_work_node_dict):
        self.runtime_ctx.update(work_node_name, update_work_node_dict, domain = "work_nodes")
        return

    def get(self, work_node_name):
        return self.runtime_ctx.get(work_node_name, domain = "work_nodes")


from .init_worker_agent import export as init_worker_agent
from .manage_context import export as manage_context
from .generate_prompt import export as generate_prompt
from .assemble_request_messages import export as assemble_request_messages
from .register_response_suffix import export as register_response_suffix
from .request import export as request


import asyncio

async def repost_delta(delta_data, **kwargs):
    listener = kwargs["listener"]
    await listener.emit("delta", delta_data)

async def repost_delta_full(delta_data, **kwargs):
    listener = kwargs["listener"]
    await listener.emit("delta_full", delta_data)

async def repost_done(done_data, **kwargs):
    listener = kwargs["listener"]
    await listener.emit("done", done_data)

async def repost_done_full(done_data, **kwargs):
    listener = kwargs["listener"]
    await listener.emit("done_full", done_data)

async def register_response_suffix(runtime_ctx, **kwargs):
    listener = kwargs["listener"]
    delta_suffix = runtime_ctx.get("delta_suffix")
    delta_full_suffix = runtime_ctx.get("delta_full_suffix")
    done_suffix = runtime_ctx.get("done_suffix")
    done_full_suffix = runtime_ctx.get("done_full_suffix")
    if callable(delta_suffix):
        delta_suffix = [delta_suffix]
    if callable(delta_full_suffix):
        delta_full_suffix = [delta_full_suffix]
    if callable(done_suffix):
        done_suffix = [done_suffix]
    if callable(done_full_suffix):
        done_full_suffix = [done_full_suffix]
    for delta_suffix_func in delta_suffix:
        async def delta_suffix_func_with_kwargs(data):
            return await delta_suffix_func(data, **kwargs)
        listener.on("extract:delta", delta_suffix_func_with_kwargs)
    for delta_full_suffix_func in delta_full_suffix:
        async def delta_full_suffix_func_with_kwargs(data):
            return await delta_full_suffix_func(data, **kwargs)
        listener.on("extract:delta_full", delta_full_suffix_func_with_kwargs)
    for done_suffix_func in done_suffix:
        async def done_suffix_func_with_kwargs(data):
            return await done_suffix_func(data, **kwargs)
        listener.on("extract:done", done_suffix_func_with_kwargs)
    for done_full_suffix_func in done_full_suffix:
        async def done_full_suffix_func_with_kwargs(data):
            return await done_full_suffix_func(data, **kwargs)
        listener.on("extract:done_full", done_full_suffix_func_with_kwargs)
    reply_handler = runtime_ctx.get("reply_handler")
    if reply_handler:
        def update_final_reply(data):
            runtime_ctx.set("final_reply", reply_handler(data))
            return
        listener.on("done", update_final_reply)

    if runtime_ctx.get("is_debug"):
        listener.on("done_full", lambda data: print("[RESPONSE]\n", data))

def export(agently):
    agently\
        .manage_work_node("register_response_suffix")\
        .set_main_func(register_response_suffix)\
        .set_runtime_ctx({
            "delta_suffix": {
                "layer": "agent",
                "alias": {
                    "set": "set_delta_suffix",
                    "get": "get_delta_suffix",
                },
                "default": [repost_delta],
            },
            "delta_full_suffix": {
                "layer": "agent",
                "alias": {
                    "set": "set_delta_full_suffix",
                    "get": "get_delta_full_suffix",
                },
                "default": [repost_delta_full],
            },
            "done_suffix": {
                "layer": "agent",
                "alias": {
                    "set": "set_done_suffix",
                    "get": "get_done_suffix"
                },
                "default": [repost_done],
            },
            "done_full_suffix": {
                "layer": "agent",
                "alias": {
                    "set": "set_done_full_suffix",
                    "get": "get_done_full_suffix"
                },
                "default": [repost_done_full],
            },
            "reply_handler": {
                "layer": "request",
                "alias": { "set": "reply" },
            },
        })\
        .register()
    return


class EnvBox(object):
    def __init__(self, agently):
        self.env_data = {
            # 目标
            "target": "",
            # 整体计划
            "overall_plan": "",
            # 结果数据
            "data": {},
            # 可调用的方法
            "func": {}
            
        }
        self.overwatcher = None
        self.agents = {}
        self.agently = agently
        self.options = {
            "show_process": False,
        }

    def set_overwatcher(self, overwatcher):
        self.overwatcher = overwatcher
        return self
    
    def set_plan(self, plan):
        self.env_data["overall_plan"] = plan
        return self

    def set_target(self, target):
        self.env_data["target"] = target
        return self

    def set_data(self, key, value):
        self.env_data["data"].update({ key: value })
        return self

    def get_data(self, key, value):
        return self.env_data["data"]["key"] if "key" in self.env_data["data"] else None

    def register_func(self, func_name, desc, parameter_requirement, func):
        self.env_data["func"].update({
            func_name:
                {
                    "desc": desc,
                    "parameter_requirement": parameter_requirement,
                    "func": func
                }
        })
        return self

    def register_agent(self, agent_name, agent_purpose, agent):
        self.agents.update({
            agent_name:
                {
                    "purpose":agent_purpose,
                    "agent": agent
                }
        })
        return self

    def set_options(self, key, value):
        self.options.update({ key: value })
        return self

    def set_main_func(self, main_func):
        self.__main_func = main_func
        return self

    def __main_func(self):
        is_solved = False
        achievement = []
        final_reply = None

        # 整理可用的 agent 信息
        agents_info = []
        for agent_name in self.agents.keys():
            agents_info.append({
                "name": agent_name,
                "purpose": self.agents[agent_name]["purpose"]
            })
        # 整理可用的 function
        func_info = []
        for func_name in self.env_data["func"].keys():
            func_info.append({
                "name": func_name,
                "desc": self.env_data["func"][func_name]["desc"],
                "parameter_requirement": self.env_data["func"][func_name]["parameter_requirement"]
            })
        overwatcher = self.overwatcher
        # 全局观察者设定
        overwatcher\
            .use_role(True)\
            .set_role("Relevant Data", self.env_data["data"])\
            .set_role("Agent You Can Call", agents_info)\
            .set_role("Func You Can Call", func_info)
        
        # 根据目标、agent角色、可调用的方法，输出整体的执行计划
        if self.env_data["overall_plan"] == "":
            overall_planner_session = overwatcher.create_session()
            overall_plan = overall_planner_session\
                    .input({
                        "achievement": achievement,
                        "target": self.env_data["target"]
                    })\
                    .output({
                        "overall_plan": ("String", "Based on the {{target}} and the {{Agent You Can Call}} and the {{Func You Can Call}}, establish your detailed global plan（Using \"ReAct\" mode）, requiring a clear description for each step indicating which role should perform what task, and strive to efficiently accomplish the target."),
                    })\
                    .start()
            if self.options["show_process"]:
                print('[Overall plan]\n', overall_plan)
            self.env_data['overall_plan'] = overall_plan
        
        next_task = "Prepare to start."

        while not is_solved:
            next_move_judgement_session = overwatcher.create_session()

            next_move_judgement = next_move_judgement_session\
                .input({
                    "achievement": achievement,
                    "target": self.env_data["target"],
                    "current_task": next_task
                })\
                .output({
                    "next_task": ("String", "Describe the next child task according {{Overall plan}} and {{current_task}}."),
                    "next_move_type": ("call | text | reply", "Judge type of your next move:'call' to call agent or func to help; 'text' to output your text anwser to {{next_task}}; 'reply' to generate final reply to {{target}} according {{Overall plan}} and {{achievement}}"),
                    "call": {
                        "type": ("agent | func | null", "output 'agent' or 'func' you want to call if {{next_move_type}} is 'call', else output 'null'"),
                        "name": ("String", "name of agent or func in {{Agent You Can Call}} or {{Func You Can Call}}"),
                        "parameter": ("String | dict according to {{parameter_requirement}}", "input string you want to tell the agent to do if {{call.type}} is 'agent'; parameters dict the func required if {{call.type}} is 'func'.")
                    },
                    "text": {
                        "type": ("text | reply", "'text' for anwser to {{next_task}} that will storage to {{achievement}}; 'reply' for final reply to {{target}}."),
                        "content": ("String",)
                    },
                })\
                .start()

            if self.options["show_process"]:
                print('[Next Move Judgement]\n', next_move_judgement)
            
            next_task = next_move_judgement["next_task"]

            # 一、处理调用逻辑
            if next_move_judgement["next_move_type"] == "call":
                # 处理 agent 调用
                if next_move_judgement["call"]["type"] == "agent":
                    agent_session = self.agents[next_move_judgement["call"]["name"]]["agent"].create_session()
                    agent_result = agent_session\
                        .input({
                            "achievement": achievement,
                            "target": next_move_judgement["call"]["parameter"],
                        })\
                        .output({
                            "output": ("String", "Your work output to {{target}}.")
                        })\
                        .start()
                    achievement.append(f"Agent [{ next_move_judgement['call']['name'] }]: { agent_result['output'] }")
                    if self.options["show_process"]:
                        print('[Achievement]\n', agent_result)
                # 处理 function 调用
                if next_move_judgement["call"]["type"] == "func":
                    func_result = self.env_data["func"][next_move_judgement["call"]["name"]]["func"](next_move_judgement["call"]["parameter"])
                    achievement.append(f"Func [{ next_move_judgement['call']['name'] }]: { func_result }")
                    if self.options["show_process"]:
                        print('[Achievement]\n', func_result)
            # 二、处理文本内容返回
            if next_move_judgement["next_move_type"] == "text":
                achievement.append(f"Res: { next_move_judgement['text']['content'] }")
                if self.options["show_process"]:
                    print('[Achievement]\n', next_move_judgement['text']['content'])
            # 三、处理最终结果
            if next_move_judgement["next_move_type"] == "reply":
                achievement.append(f"Final: { next_move_judgement['text']['content'] }")
                if self.options["show_process"]:
                    print('[Achievement]\n', next_move_judgement['text']['content'])
                final_reply = next_move_judgement["text"]["content"]
                is_solved = True

        return {
            "process_achievement": achievement,
            "final_reply": final_reply
        }

    def start(self):
        return self.__main_func()


import asyncio
import aiohttp
import json

from .loads_and_fix import loads_and_fix

'''
OpenAI GPT
'''

def prepare_request_data(runtime_ctx):
    #default request_data
    request_data = {
        "llm_url": "https://api.openai.com/v1/chat/completions",
        "data": {
            "model": "gpt-3.5-turbo",
            "temperature": 1,
        },
    }
    #check auth info
    llm_auth = runtime_ctx.get("llm_auth")
    if "GPT" in llm_auth:
        request_data.update({ "llm_auth": llm_auth["GPT"] })
    else:
        raise Exception(f"[request]: llm_auth must be set to agent. use agent.set_llm_auth('GPT', <Your-API-Key>) to set.")
    #update user customize settings
    cus_llm_url = runtime_ctx.get("llm_url")
    cus_llm_url = cus_llm_url["GPT"] if cus_llm_url and "GPT" in cus_llm_url else None
    if cus_llm_url:
        request_data.update({ "llm_url": cus_llm_url })
    request_data.update({ "proxy": runtime_ctx.get("proxy") })
    cus_request_options = runtime_ctx.get("request_options")
    cus_request_options = cus_request_options["GPT"] if cus_request_options and "GPT" in cus_request_options else None
    if isinstance(cus_request_options, dict):
        for options_name, options_value in cus_request_options.items():
            request_data["data"].update({ options_name: options_value })
    #get request messages, no transform because standard message style follow GPT API style
    request_data["data"].update({ "messages": runtime_ctx.get("request_messages") })
    return request_data

async def request(request_data, listener):
    url = request_data["llm_url"]
    data = request_data["data"].copy()
    data["stream"] = False
    headers = {
        "Authorization": f"Bearer { request_data['llm_auth'] }",
        "Content-Type": "application/json",
        "Accept": "application/json",
    }
    proxy = request_data["proxy"] if "proxy" in request_data else {}
    async with aiohttp.ClientSession() as session:
        async with session.post(url, headers=headers, json=data, proxy=proxy) as response:
            response = await response.json()
            await listener.emit("response:done", response)
            return response

async def streaming(request_data, listener):
    url = request_data["llm_url"]
    data = request_data["data"].copy()
    data["stream"] = True
    headers = {
        "Authorization": f"Bearer { request_data['llm_auth'] }",
        "Content-Type": "application/json",
        "Accept": "text/event-stream",
    }
    proxy = request_data["proxy"] if "proxy" in request_data else {}
    async with aiohttp.ClientSession() as session:
        async with session.post(url, headers=headers, json=data, proxy=proxy) as response:
            async for chunk in response.content.iter_chunks():
                if chunk[0]:
                    delta = chunk[0].decode('utf-8')[6:][:-2]
                    if delta != '[DONE]':
                        await listener.emit("response:delta", json.loads(delta))
                    else:
                        await listener.emit("response:done", None)

async def handle_response(listener, runtime_ctx, worker_agent):
    buffer = {
        "message": { "role": "", "content": "" },
        "finish_reason": None
    }    
    
    async def handle_response_delta(delta_data):
        delta_content = delta_data["choices"][0]
        await listener.emit("extract:delta_full", delta_content)
        if "delta" in delta_content and "content" in delta_content["delta"]:
            await listener.emit("extract:delta", delta_content["delta"]["content"])
        if "delta" in delta_content and delta_content["delta"] != "":
            for key, value in delta_content["delta"].items():
                value = buffer["message"][key] + value if key in buffer["message"] else value
                buffer["message"].update({ key: value })
        if "finish_reason" in delta_content and delta_content["finish_reason"] != None:
            buffer["finish_reason"] = delta_content["finish_reason"]
        return

    listener.on("response:delta", handle_response_delta)

    async def handle_response_done(done_data):
        if done_data != None:
            await listener.emit("extract:done_full", done_data["choices"][0])
            content = done_data["choices"][0]["message"]["content"]
            prompt_output_format = runtime_ctx.get("prompt_output_format")
            content = await loads_and_fix(content, prompt_output_format, worker_agent=worker_agent, is_debug=runtime_ctx.get("is_debug"))
            await listener.emit("extract:done", content)
        else:
            await listener.emit("extract:done_full", buffer)
            await listener.emit("extract:done", buffer["message"]["content"])

    listener.on("response:done", handle_response_done)


'''
TEXT
'''
def generate_output_format_text(prompt_output):
    return prompt_output

'''
JSON
'''
def transform_to_json_desc(origin, layer_count=0):
    if isinstance(origin, dict):
        json_string = ""
        if layer_count > 0:
            json_string += "\n"
        json_string += ("\t" * layer_count) + "{\n"
        for key, value in origin.items():
            json_string += ("\t" * (layer_count + 1)) + "\"" + key + "\": " + transform_to_json_desc(value, layer_count + 1) + "\n"
        if layer_count > 0:
            json_string += ("\t" * (layer_count + 1)) + "},"
        else:
            json_string += "}"
        return json_string
    elif isinstance(origin, list):
        json_string = ""
        if layer_count > 0:
            json_string += "\n"
        json_string += ("\t" * layer_count) + "[\n"
        for item in origin:
            json_string += ("\t" * (layer_count + 1)) + transform_to_json_desc(item, layer_count + 1) + ",\n"
        json_string += ("\t" * (layer_count + 1)) + "...\n"
        if layer_count > 0:
            json_string += ("\t" * layer_count) + "],"
        else:
            json_string += "]"
        return json_string
    elif isinstance(origin, tuple):
        json_string = f"<{ origin[0] }>,"
        if len(origin) >= 2:
            json_string += f"//{ origin[1] }"
        return json_string
    else:
        return str(origin)

def generate_output_format_json(prompt_output):
    return {
        "TYPE": "可解析的JSON字符串，无需解释",
        "FORMAT": transform_to_json_desc(prompt_output, 0),
    }
    '''
    return {
        "TYPE": "JSON String can be parsed in Python",
        "FORMAT": transform_to_json_desc(prompt_output, 0)
    }
    '''

'''
YAML
'''
def transform_to_yaml_desc(origin, layer_count=0):
    if isinstance(origin, dict):
        yaml_string = ""
        yaml_string += "  " * layer_count
        for key, value in origin.items():
            yaml_string += key + ": " + transform_to_yaml_desc(value, layer_count + 1) + "\n"
        return yaml_string
    elif isinstance(origin, list):
        yaml_string = ""
        if layer_count > 0:
            yaml_string += "\n"
        for item in origin:
            yaml_string += ("  " * (layer_count)) + "- " + transform_to_yaml_desc(item, layer_count) + "\n"
        yaml_string += ("  " * (layer_count)) + "- ...\n"
        return yaml_string
    elif isinstance(origin, tuple):
        yaml_string = f"<{ origin[0] }>"
        if len(origin) >= 2:
            yaml_string += f" #{ origin[1] }"
        return yaml_string
    else:
        return str(origin)

def generate_output_format_yaml(prompt_output):
    return {
        "TYPE": "可解析的YAML字符串，无需解释，不要包含可能会引起解析混淆的特殊字符",
        "FORMAT": "```yaml\n" + transform_to_yaml_desc(prompt_output, 0) + "```\n",
        "复查": "请确保{{FORMAT}}提到的所有字段都输出了",
    }
    '''
    return {
        "TYPE": "YAML, allow_unicode=True, can be parsed in Python by yaml.safe_load() without explanation and decoration.",
        "FORMAT": "```yaml\n" + transform_to_yaml_desc(prompt_output, 0) + "```\n",
        "DOUBLE CHECK": "Make sure every key {{FORMAT}} required is in {{OUTPUT}}"
    }
    '''

'''
PROMPT Structure
'''
def generate_prompt_structure_default(request_prompt_dict, layer_count = 1):
    request_prompt = ""
    for key, content in request_prompt_dict.items():
        request_prompt += f"{ '#' * layer_count } { key }:\n"
        if isinstance(content, dict):
            request_prompt += generate_prompt_structure_default(content, layer_count + 1) + '\n'
        else:
            request_prompt += str(content) + '\n'
        if layer_count == 1:
            request_prompt += "\n"
    if layer_count == 1:
        request_prompt += "# OUTPUT:\n"
    return request_prompt

def update_process(work_node_management):
    result = work_node_management\
        .set_process("generate_output_format", generate_output_format_json, "JSON")\
        .set_process("generate_output_format", generate_output_format_yaml, "YAML")\
        .set_process("generate_output_format", generate_output_format_text, "TEXT")\
        .set_process("generate_prompt_structure", generate_prompt_structure_default, "default")\
        .set_runtime_ctx({
            "prompt_output_format": {
                "layer": "request",
                "alias": { "set": "format" },
                "default": "TEXT",
            },
            "prompt_style": {
                "layer": "session",
                "alias": { "set": "style" },
                "default": "default",
            },
        })\
        .update()
    return


import asyncio
import json

def shorten_context(context_messages, context_length):
    short_context_messages = context_messages
    if len(json.dumps(short_context_messages)) <= context_length:
        return short_context_messages
    else:
        short_context_messages = short_context_messages[1:]
        return shorten_context(short_context_messages, context_length)

def generate_context(runtime_ctx, **kwargs):
    session_runtime_ctx = kwargs["session_runtime_ctx"]
    full_context_messages = session_runtime_ctx.get("full_context_messages")
    if full_context_messages == None:
        session_runtime_ctx.set("full_context_messages", [])
        full_context_messages = []
    request_context_messages = session_runtime_ctx.get("request_context_messages")
    if request_context_messages == None:
        session_runtime_ctx.set("request_context_messages", [])
        request_context_messages = []
    extend_context_messages = runtime_ctx.get("extend_context_messages")
    context_length = runtime_ctx.get("context_length")
    if extend_context_messages:
        full_context_messages.extend(extend_context_messages)
        request_context_messages.extend(extend_context_messages)
    use_context = runtime_ctx.get("use_context")
    if use_context:
        request_context_messages = shorten_context(request_context_messages, context_length)
        session_runtime_ctx.set("request_context_messages", request_context_messages)
    else:
        session_runtime_ctx.set("request_context_messages", extend_context_messages)
    session_runtime_ctx.set("full_context_messages", full_context_messages)
    return

async def manage_context(runtime_ctx, **kwargs):
    session_runtime_ctx = kwargs["session_runtime_ctx"]

    generate_context(runtime_ctx, **kwargs)
    
    async def save_context_after_request(data):
        prompt_input = runtime_ctx.get("prompt_input")
        prompt_instruct = runtime_ctx.get("prompt_instruct")
        prompt_output = runtime_ctx.get("prompt_output")
        if prompt_input:
            request_content = str(prompt_input)
        elif prompt_output:
            request_content = str(prompt_output)
        else:
            request_content = str(rompt_instruct)
        final_reply = runtime_ctx.get("final_reply")
        new_context_messages = [
            { "role": "user", "content": str(request_content) },
            { "role": "assistant", "content": str(data) }
        ]
        session_runtime_ctx.extend("full_context_messages", new_context_messages)
        session_runtime_ctx.extend("request_context_messages", new_context_messages)

    listener = kwargs["listener"]
    listener.on("workflow_finish", save_context_after_request)

def export(agently):
    agently\
        .manage_work_node("manage_context")\
        .set_main_func(manage_context)\
        .set_runtime_ctx({
            "use_context": {
                "layer": "agent",
                "alias": { "set": "use_context" },
            },
            "extend_context_messages": {
                "layer": "request",
                "alias": {
                    "extend": "extend_context",
                },
                "default": [],
            },
            "context_length": {
                "layer": "agent",
                "alias": {
                    "set": "set_context_length",
                },
                "default": 3500,
            }
        })\
        .register()
    return


from .RuntimeCtx import RuntimeCtx
from .WorkNodes import WorkNodes
from .Workflows import Workflows
from .runtime_ctx_settings import inject_alias, set_default_runtime_ctx

class Blueprint(object):
    def __init__(self, agently):
        self.runtime_ctx = RuntimeCtx(agently.runtime_ctx)
        self.work_nodes = WorkNodes(self.runtime_ctx)
        self.workflows = Workflows(self.runtime_ctx)

    def manage_work_node(self, work_node_name):
        return self.work_nodes.manage_work_node(work_node_name)

    def set_workflow(self, workflow, **kwargs):
        workflow_name = kwargs.get("name", "default")
        return self.workflows.set(workflow_name, workflow)

    def set(self, key, value, **kwargs):
        self.runtime_ctx.set(key, value, **kwargs)
        return self

    def use(self, inject_funcs):
        if not isinstance(inject_funcs, list):
            if not callable(inject_funcs):
                raise Exception("[Agently use]: .use() must pass on a function list or a function.")
            else:
                inject_funcs = [inject_funcs]
        for inject_func in inject_funcs:
            if not callable(inject_func):
                raise Exception("[Agently use]: .use() must pass on a callable function.")
            else:
                inject_func(self)
        return self

    def append(self, key, value, **kwargs):
        self.runtime_ctx.append(key, value, **kwargs)
        return self

    def init(self):
        inject_alias(self, "agent")
        set_default_runtime_ctx(self.runtime_ctx, "agent")
        return


import asyncio
import aiohttp
import json

from .loads_and_fix import loads_and_fix

'''
MiniMax
'''

def prepare_request_data(runtime_ctx):
    #default request_data
    request_data = {
        "llm_url": "https://api.minimax.chat/v1/text/chatcompletion",
        "data": {
            "model": "abab5.5-chat",
            "temperature": 1,
        },
    }
    #check auth info
    llm_auth = runtime_ctx.get("llm_auth")
    if "MiniMax" in llm_auth:
        request_data.update({ "llm_auth": llm_auth["MiniMax"] })
    else:
        raise Exception("[request]: llm_auth must be set to agent. use agent.set_llm_auth('MiniMax', { 'group_id': '...', 'api_key': '...' }) to set.")
    #update user customize settings
    cus_llm_url = runtime_ctx.get("llm_url")
    cus_llm_url = cus_llm_url["MiniMax"] if cus_llm_url and "MiniMax" in cus_llm_url else None
    if cus_llm_url:
        request_data.update({ "llm_url": cus_llm_url })
    request_data.update({ "proxy": runtime_ctx.get("proxy") })
    cus_request_options = runtime_ctx.get("request_options")
    cus_request_options = cus_request_options["MiniMax"] if cus_request_options and "MiniMax" in cus_request_options else None
    if isinstance(cus_request_options, dict):
        for options_name, options_value in cus_request_options.items():
            request_data["data"].update({ options_name: options_value })
    #get request messages
    request_messages = runtime_ctx.get("request_messages")
    #transform messages format
    prompt = ''
    minimax_user_name = runtime_ctx.get("minimax_user_name")
    minimax_bot_name = runtime_ctx.get("minimax_bot_name")
    final_request_messages = []
    for message in request_messages:
        if message["role"] == "system":
            prompt += f"{ message['content'] }\n"
        else:
            role_mapping = {
                "assistant": "BOT",
                "user": "USER",
            }
            final_request_messages.append({
                "sender_type": role_mapping[message["role"]],
                "text": message["content"]
            })
    request_data["data"].update({
        "messages": final_request_messages,
    })
    if prompt != '':
        request_data["data"].update({
            "prompt": prompt,
            "role_meta": {
                "user_name": minimax_user_name if minimax_user_name else "USER",
                "bot_name": minimax_bot_name if minimax_bot_name else "BOT",
            },
        })
    return request_data

async def request(request_data, listener):
    url = request_data["llm_url"] + "?GroupId=" + request_data["llm_auth"]["group_id"]
    data = request_data["data"].copy()
    data["stream"] = False
    headers = {
        "Authorization": f"Bearer { request_data['llm_auth']['api_key'] }",
        "Content-Type": "application/json",
    }
    proxy = request_data["proxy"] if "proxy" in request_data else {}
    async with aiohttp.ClientSession() as session:
        async with session.post(url, headers=headers, json=data, proxy=proxy) as response:
            response = await response.text()
            response = json.loads(response)
            if response["base_resp"]["status_code"] != 0:
                raise Exception(response["base_resp"])
            await listener.emit("response:done", response)
            return response

async def streaming(request_data, listener):
    url = request_data["llm_url"] + "?GroupId=" + request_data["llm_auth"]["group_id"]
    data = request_data["data"].copy()
    data["stream"] = True
    data["use_standard_sse"] = True
    headers = {
        "Authorization": f"Bearer { request_data['llm_auth']['api_key'] }",
        "Content-Type": "application/json",
        "Accept": "text/event-stream",
    }
    proxy = request_data["proxy"] if "proxy" in request_data else {}
    #response = requests.post(url, headers=headers, json=data, stream=True ,proxies=proxies)
    #return response
    async with aiohttp.ClientSession() as session:
        async with session.post(url, headers=headers, json=data, proxy=proxy) as response:
            async for chunk in response.content.iter_chunks():
                if chunk[0]:
                    delta = json.loads(chunk[0].decode('utf-8')[6:][:-2])
                    if delta["reply"] == "":
                        await listener.emit("response:delta", delta)
                    else:
                        await listener.emit("response:done", delta)

async def handle_response(listener, runtime_ctx, worker_agent):
    is_streaming = runtime_ctx.get("is_streaming")

    if is_streaming:
        async def handle_response_delta(delta_data):
            delta_content = delta_data["choices"][0]
            if "extract:delta" in delta_content and delta_content["extract:delta"] != "":
                await listener.emit("extract:delta_full", { "extract:delta": { "content": delta_content["extract:delta"] }, "finish_reason": None })
                await listener.emit("extract:delta", delta_content["extract:delta"])
            if "finish_reason" in delta_content:
                await listener.emit("extract:delta_full", { "extract:delta": { "content": "", "finish_reason": delta_content["finish_reason"] } })

        listener.on("response:delta", handle_response_delta)
        await listener.emit("extract:delta_full", { "extract:delta": { "role": "assistant", "content": "" }, "finish_reason": None })

        async def handle_response_done(done_data):
            await listener.emit("extract:done_full", {
                "index": done_data["choices"][0]["index"],
                "message": { "role": "assistant", "content": done_data["reply"] },
                "finish_reason": done_data["choices"][0]["finish_reason"],
            })
            content = done_data["reply"]
            prompt_output_format = runtime_ctx.get("prompt_output_format")
            if prompt_output_format.upper() in ("JSON", "YAML"):
                content = await loads_and_fix(content, prompt_output_format, worker_agent=worker_agent, is_debug=runtime_ctx.get("is_debug"))
            await listener.emit("extract:done", content)

        listener.on("response:done", handle_response_done)
    else:
        async def handle_response_done(done_data):
            done_content = done_data["choices"][0]
            await listener.emit("extract:done_full", {
                "index": done_content["index"],
                "message": { "role": "assistant", "content": done_content["text"] },
                "finish_reason": done_content["finish_reason"],
            })
            content = done_content["text"]
            prompt_output_format = runtime_ctx.get("prompt_output_format")
            content = await loads_and_fix(content, prompt_output_format, worker_agent=worker_agent, is_debug=runtime_ctx.get("is_debug"))
            await listener.emit("extract:done", content)

        listener.on("response:done", handle_response_done)